/*
** Module   : ClientCore.java
** Abstract : core client loop which handles init and connects to server
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------------------Description---------------------------------
** 001 GES 20110928 Moved core logic here from ClientDriver so that the same
**                  code could be used from multiple callers.
** 002 GES 20111028 Added support for startup of a client session within the
**                  same process as the serve.
** 003 GES 20111111 Pass single mode flag to thin client.
** 004 CA  20130529 If this is authenticated as an appserver, register the needed server exports.
** 005 CA  20130704 Removed H004, as the appserver launching was reworked.
** 006 CA  20131004 Expose client-side parameters to the server-side application code.
** 007 EVL 20131104 Batch mode support.
*/

package com.goldencode.p2j.main;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;                          
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.util.*;

import com.goldencode.p2j.ui.*; // EVL***
import com.goldencode.p2j.ui.client.chui.driver.batch.*;// EVL***

/**
 * Core looping logic that performs initialization of a client UI and
 * establishes a session with the server.
 * <p>
 * This code provides supports the startup of a client session within the
 * same process as the server.  This can be used to duplicate the traditional
 * Progress 4GL fat client approach (a "single process" mode). It can also be
 * used to implement multiple client sessions within a true server process
 * which would be used to support web-based clients.
 */
public class ClientCore
{
   /**
    * Start the client loop.  This will not return until the user is exiting
    * the client. The core loop initializes logging, the security manager,
    * the UI and then tries to connect to the server. During the connection
    * attempt the server will send down a login screen or if guest access is
    * enabled, the login will be bypassed. Once login has occurred (or been
    * bypassed), the session is established and the directory manager and the
    * UI client finish initialization.  Finally the main application entry
    * point is called on the server.  The server will not return from that call
    * until the user asks to quit, a CTRL-C occurs (stop condition) or some
    * abnormal end occurs.  Any of these cases cause the loop to terminate and
    * this method will then return.
    * <p>
    * This method implements the full, normal client JVM startup and cannot be
    * used within the server process.
    * 
    * @param    config
    *           The configuration to use for setup of the client and for the
    *           server connection.
    * @param    file
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    * @param    args
    *           Application-specific command line arguments. Ignored at this
    *           time.
    */
   public static void start(BootstrapConfig config, boolean file, String[] args)
   throws Exception
   {
      start(config, file, false, args);
   }
   
   /**
    * Start the client loop.  This will not return until the user is exiting
    * the client. The core loop initializes logging, the security manager,
    * the UI and then tries to connect to the server. During the connection
    * attempt the server will send down a login screen or if guest access is
    * enabled, the login will be bypassed. Once login has occurred (or been
    * bypassed), the session is established and the directory manager and the
    * UI client finish initialization.  Finally the main application entry
    * point is called on the server.  The server will not return from that call
    * until the user asks to quit, a CTRL-C occurs (stop condition) or some
    * abnormal end occurs.  Any of these cases cause the loop to terminate and
    * this method will then return.
    * 
    * @param    config
    *           The configuration to use for setup of the client and for the
    *           server connection.
    * @param    file
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    * @param    single
    *           <code>true</code> to startup within the server process which
    *           must bypass the shared infrastructure initialization. Use
    *           <code>false</code> for the normal client JVM startup.
    * @param    args
    *           Application-specific command line arguments. Ignored at this
    *           time.
    */
   public static void start(BootstrapConfig config,
                            boolean         file,
                            boolean         single,
                            String[]        args)
   throws Exception
   {
      boolean dbg     = config.getBoolean("logging", "debug", "enable", false);
      boolean running = true;
      
      // save off the UI driver so the same instance can be used across all
      // iterations of the core loop
      ScreenDriver driver = null;
      
      while (running)
      {
         Object          context = null; 
         Session         session = null;
         MainEntry       main    = null;
         SessionManager  sessMgr = null;
         SecurityManager sm      = null;
         
         // EVL***<<
         Thread.sleep(5000);
         // EVL***<<
         try
         {
            if (!single)
            {
               // optionally initialize detailed logging (only use this for
               // debugging sessions!); this must be done before the SM inits
               // to ensure that the logging goes to the right place
               if (dbg)
                  initLogging(file);
               
               // instantiate security manager
               sm = SecurityManager.createInstance(config);
               
               if (dbg)
                  sm.setDebugLevel(SecurityManager.DATA);
               
               // instantiate network session manager
               sessMgr = SessionManagerFactory.createLeafNode(config);
            }
            
            // initialize core UI services
            ThinClient.initializePrep(config, driver, single);
            
            // at this point it is safe to query the driver and save it off
            // in case we need to re-initialize (the same driver must be
            // re-used even though we create new instances of pretty much all
            // other UI classes)
            if (driver == null)
            {
               driver = OutputManager.getDriver();
            }
            
            if (single)
            {
               sm = SecurityManager.getInstance();
               
               // authenticate and create a security context for this thread
               context = sm.authenticateSingle(config);
            }
            else
            {
               // establish or re-establish a session with the P2J server
               // (this handles authentication too)
               session = sessMgr.connectDirect(new InterruptedExceptionHandler(),
                                               null);
                                               
               // some client code must read configuration from the directory
               DirectoryManager.initRemote(session);
            }
         }
         
         catch (Exception exc)
         {
            // STOP conditions don't need to be reported, they are an expected
            // mechanism to terminate
            if (!isCauseStop(exc))
            {
               exc.printStackTrace(System.err);
               System.err.flush();
            }
            
            break;
         }

         // finish initializing the thin client
         ThinClient.initializePost(session, context);
         
         // EVL***<<
         if (session != null)
         {
            OutputManager.initErrorWriter();
         }
         // EVL***>>
         
         // obtain the entry point
         if (single)
         {
            // client code is "right" side by convention
            main = (MainEntry)
               RemoteObject.obtainLocalInstance(MainEntry.class, false);
         }
         else
         {
            main = (MainEntry)
               RemoteObject.obtainNetworkInstance(MainEntry.class, session);
         }
   
         // determine any client-side parameters
         ClientParameters params = new ClientParameters();
         // check if it has a custom window-system set 
         String windowSystem = config.getString("client", "driver", "window-system", null);
         if (windowSystem == null)
         {
            // else, default to the driver's window system
            windowSystem = driver.getWindowSystem();
         }
         params.setWindowSystem(windowSystem);
                  
         // above this point, any CTRL-C processing that causes a STOP
         // condition to be thrown will cause an exit (this is very timing
         // sensitive but if you hold down the CTRL-C keys AND the server
         // environment is configured to force a relogin (the return from
         // standardEntry() is true) then at some point you will see such
         // an exception; we can protect against this, BUT this same behavior
         // has been seen in the 4GL, so it has been left here on purpose
            
         // EVL***<<
         if (driver != null && driver instanceof BatchDriver)
         {
            ((BatchPrimitives)driver.getPrimitives()).activate();
         }
         // EVL***>>
         // call the application
         running = main.standardEntry(params);

         // shut down session
         if (single)
         {
            ThinClient.killDependentThreads();
            sm.terminateSingle(context);
         }
         else
         {
            session.terminate();
         }
      }
   }
   
   /**
    * Inspect the chain of exceptions and report if any of them are STOP
    * conditions.
    *
    * @param    chained
    *           The top of the chain of exceptions.
    *
    * @return   <code>true</code> if any of the exceptions in the chain are
    *           STOP conditions.
    */
   public static boolean isCauseStop(Throwable chained)
   {
      boolean isStop = false;
      
      while (chained != null)
      {
         if (chained instanceof StopConditionException)
         {
            isStop = true;
            break;
         }
         
         chained = chained.getCause();
      }
      
      return isStop;
   }
   
   /**
    * Initialize logging at a detailed level for the infrastructure packages.
    *
    * @param    toFile
    *           <code>true</code> to direct logging to file. <code>false</code>
    *           for console logging.
    */
   private static void initLogging(boolean toFile)
   {
      LinkedHashMap<String, Level> loggingLevels =
         new LinkedHashMap<String, Level>();
         
      loggingLevels.put("com.goldencode.p2j.net", Level.FINEST);
      loggingLevels.put("com.goldencode.p2j.main", Level.FINEST);
      
      LogHelper.initialize(null,
                           Level.FINEST,                       // default level
                           !toFile ? Level.FINEST : Level.OFF, // console
                           toFile  ? Level.FINEST : Level.OFF, // file
                           "client_dbg_%g.log",
                           -1,
                           -1,
                           false,
                           loggingLevels);
   }
}
