/*
** Module   : OutputManager.java
** Abstract : Provides a generic interface to low level UI primitives.
**
** Copyright (c) 2010-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 GES 20100817 Cleaned up formatting, fixed/added javadoc, removed
**                  excessive calls to getters in invariant cases.
** 004 SIY 20101108 Make sure that driver will be initialized only 
**                  once even if OutputManager instances are created
**                  again and again.
** 005 SIY 20101202 Removed unused method ans slightly changed
**                  interaction with the driver in order to improve
**                  decoupling between UI and driver. Reorganized
**                  constants.
** 006 SIY 20110112 Refactoring and simplification, changes in color
**                  handling.
** 007 GES 20110921 Support for multiple drivers.
** 008 GES 20111012 Moved to context-local instead of singleton. This allows
**                  multiple client sessions to be hosted in the server.
** 009 SIY 20111101 Updated imports, minor cleanups/refactorings.
** 010 GES 20111103 Modifications to allow the screen driver instance to be
**                  obtained and reused upon re-initialization (this happens
**                  during CTRL-C processing).
** 011 SIY 20111202 Updated imports.
** 012 SIY 20111221 Added accessor for widget factory.
** 013 EVL 20130724 Replacing CHARVA with our own console driver.
** 014 EVL 20130822 Adding new 6 characters into consideration to handle comb-box and
**                  selection-list widgets drawing.
** 015 MAG 20131114 Add chui web screen driver support.
**                  The swing_chui_applet is no longer used.
** 016 EVL 20131104 Batch mode support.
*/

package com.goldencode.p2j.ui.client;

import java.util.List;
import java.lang.reflect.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.chui.driver.swing.*;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.ui.client.chui.driver.batch.*;// EVL***
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;// EVL***

/**
 * Provides a generic interface to low level user interface primitives.
 */
public class OutputManager
{
   /** CHUI native driver name. */
   public static final String CHUI_DRIVER_CONSOLE = "chui_native";
   
   /** CHUI swing frame driver name. */
   public static final String CHUI_DRIVER_SWING_FRAME = "swing_chui_frame";
     
   /** CHUI web driver name. */
   public static final String CHUI_DRIVER_WEB = "chui_web";

   // EVL***>>
   public static final String CHUI_DRIVER_BATCH = "chui_batch";
   // EVL***>>
   
   /** Bottom tee drawing char. */
   public static final char UACS_BTEE = '\u2534';

   /** Top tee drawing char. */
   public static final char UACS_TTEE = '\u252c';

   /** Lower left corner drawing char. */
   public static final char UACS_LLCORNER = '\u2514';

   /** Lower right corner drawing char. */
   public static final char UACS_LRCORNER = '\u2518';

   /** Upper right corner drawing char. */
   public static final char UACS_URCORNER = '\u2510';

   /** Upper left corner drawing char. */
   public static final char UACS_ULCORNER = '\u250c';

   /** Horizontal line drawing char. */
   public static final char UACS_HLINE = '\u2500';

   /** Vertical line drawing char. */
   public static final char UACS_VLINE = '\u2502';
   
   /** Up arrow char for vertical scrollbars. */
   public static final char UACS_UARROW = '\u2191';

   /** Down arrow char for vertical scrollbars. */
   public static final char UACS_DARROW = '\u2193';

   /** Left arrow char for horizontal scrollbars. */
   public static final char UACS_LARROW = '\u2190';

   /** Right arrow char for horizontal scrollbars. */
   public static final char UACS_RARROW = '\u2192';

   /** Checker board char for scrollbars body. */
   public static final char UACS_CKBOARD = '\u2592';

   /** Current selection char for selection lists. */
   public static final char UACS_SELECT = '\u2219'; // This is Unicode bullet char

   /** Context local instance of this class. */
   private static final ContextLocal<OutputManager> local =
      new ContextLocal<OutputManager>();
   
   /** Flag to enable debug messages. */
   protected boolean debug = false;

   /** Default (direct) drawing routines. */
   protected OutputPrimitives defaultOutput;

   /** Current drawing routines. */
   protected OutputPrimitives output;

   /** The state of invalidation. */
   protected int invalidate = 0;

   /** The screen state, which keeps track of all the clipped regions. */
   protected ScreenBitmap bitmap = null;

   /** Track the current cursor status. */
   protected boolean cursorInvalid  = false;

   /** Direct drawing routines. */
   protected OutputPrimitives directOutput;

   /** Reflects drawing in progress to avoid recursion. */
   protected boolean drawing = false;

   /** Active screen driver. */
   private ScreenDriver driver;
   
   /** Batch mode client override flag */
   private boolean batchModeOverride = false;

   /**
    * Private constructor to ensure a singleton pattern.
    *
    * @param    driver
    *           Implements low level input/output primitives.  Must NOT be
    *           <code>null</code>.
    */
   private OutputManager(ScreenDriver driver)
   {
      this.driver = driver;
   }

   /**
    * Initialize driver. This MUST be the first method called in this class.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * @param    driver
    *           Implements low level input/output primitives.  This MAY be
    *           <code>null</code> on the first invocation BUT it MUST NOT
    *           be <code>null</code> on re-initializations.
    *           
    * @throws   InterruptedException 
    * @throws   InvocationTargetException 
    */
   public synchronized static void init(BootstrapConfig cfg,
                                        ScreenDriver    driver)
   throws InterruptedException,
          InvocationTargetException
   {
      OutputManager    mgr    = newInstance(cfg, driver);
      OutputPrimitives output = mgr.driver.getPrimitives();

      mgr.output        = output;
      mgr.defaultOutput = output;
      mgr.directOutput  = output;
      mgr.bitmap        = output.getScreenBitmap();
      // EVL***
      mgr.batchModeOverride = isBatchModeForced(cfg);
      // EVL***
   }

   // EVL***<<
   private static boolean isBatchModeForced(BootstrapConfig cfg)
   {
      String driverName = cfg.getString("client", "driver", "type", CHUI_DRIVER_CONSOLE);
      
      return cfg.getBoolean("client", "mode", "batch", false) ||
             (driverName == null ? false : driverName.equalsIgnoreCase(CHUI_DRIVER_BATCH));
      
   }
   // EVL***>>
   
   /**
    * Force construction of a new instance.
    *
    * @param    cfg
    *           The bootstrap configuration file.
    * @param    driver
    *           Implements low level input/output primitives.  This MAY be
    *           <code>null</code> on the first invocation BUT it MUST NOT
    *           be <code>null</code> on re-initializations.
    * 
    * @return   The newly created instance.
    * 
    * @throws   InterruptedException 
    * @throws   InvocationTargetException 
    */
   public static synchronized OutputManager newInstance(BootstrapConfig cfg,
                                                        ScreenDriver    driver)
   throws InterruptedException,
          InvocationTargetException
   {
      if (driver == null) 
      {
         String id = cfg.getString("client",
                                   "driver",
                                   "type",
                                   CHUI_DRIVER_CONSOLE);
         
         // EVL***<<
         boolean batchMode = cfg.getBoolean("client", "mode", "batch", false); 
         
         if (batchMode || id.equalsIgnoreCase(CHUI_DRIVER_BATCH))
         {
            driver = new BatchDriver();
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_CONSOLE))
         // EVL***>>
// EVL***         if (id.equalsIgnoreCase(CHUI_DRIVER_CONSOLE))
         {
            driver = new ConsoleDriver();
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_SWING_FRAME))
         {
            TerminalOptions options = SwingChuiDriver.convertToOptions(cfg);
            final ChuiSimulator sim = ChuiClient.load(options);
            
            driver = new SwingChuiDriver(sim);
         }
         else if (id.equalsIgnoreCase(CHUI_DRIVER_WEB))
         {
            driver = new ChuiWebDriver(cfg);
         }
         else
         {
            throw new RuntimeException("Unknown driver " + id);
         }

         driver.init();
      }
      
      OutputManager mgr = new OutputManager(driver);
      local.set(mgr);
      
      return mgr;
   }

   /**
    * Get the context-specific instance, if it exists. In a single-client
    * JVM environment, this is equivalent to a singleton.
    * 
    * @return   The instance or <code>null</code> if none exists.
    */
   public static synchronized OutputManager instance()
   {
      return local.get();
   }
   
   /**
    * Access the current UI driver instance.
    *
    * @return   The current driver or <code>null</code> if the output manager
    *           is not initialized.
    */
   public static synchronized ScreenDriver getDriver()
   {
      OutputManager mgr = instance();
      
      return mgr == null ? null : mgr.driver;
   }

   // EVL***<<
   /**
    * Initializes the proper error writer handler.
    */
   public static synchronized void initErrorWriter()
   {
      OutputManager mgr = instance();
      
      if (EnvironmentOps.isBatchMode().booleanValue() ||
          (mgr != null && mgr.isBatchModeOverride()))
      {
         ErrorManager.initErrorWriter(new ErrorWriterBatch());
      }
      else
      {
         ErrorManager.initErrorWriter(new ErrorWriterInteractive());
      }
   }

   public boolean isBatchModeOverride()
   {
      return batchModeOverride;
   }
   // EVL***>>
   
   /**
    * Set cursor status (ON/OFF).
    * 
    * @param   on
    *          New cursor state.
    * 
    * @return  <code>true</code> if operation was successful.
    * 
    * @throws  IllegalStateException
    */
   public boolean setCursorStatus(boolean on) throws IllegalStateException
   {
      return driver.setCursorStatus(on);
   }

   /**
    * Clear screen.
    */
   public void clearScreen()
   {
      driver.clear();
   }

   /**
    * Generate beep signal.
    */
   public void beep()
   {
      driver.beep();
   }

   /**
    * Check if driver supports color output.
    * 
    * @return <code>true</code> if colors are supported.
    */
   public boolean hasColors()
   {
      return driver.hasColors();
   }

   /**
    * Read single keystroke from driver.
    * 
    * @return key code.
    */
   public int readKey()
   {
      return driver.readKey();
   }

   /**
    * Reset current screen mode.
    */
   public void resetMode()
   {
      driver.resetMode();
   }

   /**
    * Suspend output.
    */
   public void suspend()
   {
      driver.suspend();
   }

   /**
    * Resume normal driver operation.
    */
   public void resume()
   {
      driver.resume();
   }

   /**
    * Change terminal type to specified one.
    * 
    * @param    name
    *           New terminal type name.
    */
   public void setTerminalType(String name)
   {
      driver.setTerminalType(name);
   }

   /**
    * Get terminal type.
    * 
    * @return terminal type string.
    */
   public String terminalType()
   {
      return driver.terminalType();
   }

   /**
    * Redirect terminal output to the drawing worker represented by
    * <code>newOps</code> (or back to the interactive terminal if
    * <code>newOps</code> is <code>null</code>) and return the instance
    * representing the current output destination. This may be another
    * <code>null</code> if the interactive terminal is in use.
    * 
    * @param    newOps
    *           The drawing worker or <code>null</code> to remove redirections
    *           and activate the interactive terminal.
    * 
    * @return   The previous drawing worker instance.
    */
   public OutputPrimitives switchOutput(OutputPrimitives newOps)
   {
      OutputPrimitives old = output;

      if (newOps == null)
      {
         output = defaultOutput;
      }
      else
      {
         output = newOps;
      }

      bitmap = output.getScreenBitmap();

      return old;
   }

   /**
    * Reports if the terminal is currently redirected to a stream.
    * 
    * @return <code>true</code> if the terminal is redirected to a stream,
    *         <code>false</code> if the terminal is interactive.
    */
   public boolean isRedirected()
   {
      return !output.equals(defaultOutput);
   }

   /**
    * Obtain the direct output worker.
    * 
    * @return The output worker used to draw directly to the screen with
    *         ncurses
    */
   public OutputPrimitives getDirectOutput()
   {
      return directOutput;
   }

   /**
    * Replace direct output implementation.
    * 
    * @param    direct
    *           New direct output primitives implementation.
    */
   public void setDirectOutput(OutputPrimitives direct)
   {
      if (direct != null)
         directOutput = direct;
   }

   /**
    * Set the interactive output worker. This method should be used only to
    * provide a replacement interactive output driver. Use switchOutput()
    * method to redirect output to a stream and back.
    * 
    * @param    newOutput
    *           The output worker to be used by the interactive terminal.
    */
   public void setInteractiveOutput(OutputPrimitives newOutput)
   {
      defaultOutput = newOutput;
      output = newOutput;

      bitmap = output.getScreenBitmap();
   }

   /**
    * Obtain the interactive output worker.
    * 
    * @return The output worker used by the interactive terminal.
    */
   public OutputPrimitives getInteractiveOutput()
   {
      return defaultOutput;
   }

   /**
    * Append character with specified color at current cursor location and 
    * advance the cursor.
    *  
    * @param   chr
    *          Character to append.
    * @param   color
    *          Color and attribute.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager append(int chr, Color color)
   {
      if (!cursorInvalid && bitmap.canDraw(cursor()))
      {
         output.append(chr, color);
      }
      else
      {
         cursorInvalid = !output.cursorAt(output.getX() + 1, output.getY());
      }
      
      return this;
   }

   /**
    * Draw horizontal line.
    * 
    * @param   len
    *          Line length.
    * @param   color
    *          Line color.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager hLine(int len, Color color)
   {
      if (cursorInvalid)
      {
         return this;
      }

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.hLine(len, color);
         return this;
      }

      Point p = cursor();

      // split the line into drawable fragments and draw them
      List<Point> frags = bitmap.getHorizontalFragments(p, len);
      
      for (Point frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         at(p.x + start, p.y);
         output.hLine(length, color);
      }
      
      return this;
   }

   /**
    * Append string with specified color at current cursor location and 
    * advance the cursor.
    *  
    * @param   text
    *          Character to append.
    * @param   color
    *          Color and attribute.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager append(String text, Color color)
   {
      if (cursorInvalid)
      {
         return this;
      }

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.append(text, color);
         return this;
      }

      Point p = cursor();

      for (Point frag : bitmap.getHorizontalFragments(p, text.length()))
      {
         at(p.x + frag.x, p.y);
         output.append(text.substring(frag.x, frag.y + 1), color);
      }

      // reposition the cursor to the first position behind the string
      // no matter how the clipping went
      return at(p.x + text.length(), p.y);
   }

   /**
    * Draw a vertical line of the specified length starting at the current
    * cursor position.
    * 
    * @param   len
    *          Vertical line length.
    * @param   color
    *          Line color.
    *          
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager vLine(int len, Color color)
   {
      if (cursorInvalid)
      {
         return this;
      }

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.vLine(len, color);
         return this;
      }

      Point p = cursor();

      // split the line into drawable fragments and draw them
      List<Point> frags = bitmap.getVerticalFragments(p, len);
      
      for (Point frag : frags)
      {
         at(p.x, p.y + frag.x);
         output.vLine(frag.y - frag.x + 1, color);
      }
      
      return this;
   }

   /**
    * Clear specified area of the screen.
    * 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    */
   public void clear(Point origin, Dimension size, Color color)
   {
      Rectangle rect = new Rectangle(origin, size);
      if (bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return;
         }

         if (debug)
         {
            System.err.println("clear: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(rect);

         if (rect == null)
         {
            return;
         }
      }

      int left   = rect.left();
      int top    = rect.top();
      int bottom = rect.bottom();
      int right  = rect.right();
      
      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.clearArea(left, top, right, bottom, color);
         return;
      }

      List<Point> frags = null;
      boolean horizontal = size.width >= size.height;
      int len    = horizontal ? size.width : size.height;
      int begin  = horizontal ? top : left;
      int finish = horizontal ? bottom : right;

      for (int i = begin; i <= finish; i++)
      {
         if (horizontal)
         {
            frags = bitmap.getHorizontalFragments(i, left, len);
         }
         else
         {
            frags = bitmap.getVerticalFragments(top, i, len);
         }

         for (Point frag : frags)
         {
            int start = frag.x;
            int end = frag.y;
            int length = end - start + 1;

            int y = 0;
            int x = 0;
            int height = 0;
            int width = 0;

            if (horizontal)
            {
               y = i;
               height = 1;
               x = left + start;
               width = length;
            }
            else
            {
               y = top + start;
               height = length;
               x = i;
               width = 1;
            }

            output.clearArea(x, y, x + width - 1, y + height - 1, color);
         }
      }
   }

   /**
    * Draw box at specified location with provided size and color.
    * 
    * @param   origin
    *          Upper left corner of the box.
    * @param   size
    *          Box dimensions.
    * @param   color
    *          Color and attribute of the box lines.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager box(Point origin, Dimension size, Color color)
   {
      return box(new Rectangle(origin, size), color);
   }

   /**
    * Obtain border color for specified widget.
    * 
    * @param   widget
    *          Widget for which border color should be obtained.
    *          
    * @return  Widget border color.
    */
   public Color borderColor(Widget widget)
   {
      return output.borderColor(widget);
   }

   /**
    * Draw box at specified location with provided size and color.
    * 
    * @param   rect
    *          The box to draw.
    * @param   color
    *          Color and attribute of the box lines.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   private OutputManager box(Rectangle rect, Color color)
   {
      if (bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return this;
         }

         if (debug)
         {
            System.err.println("box: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(rect);

         if (rect == null)
         {
            return this;
         }
      }

      int left   = rect.left();
      int top    = rect.top();
      int bottom = rect.bottom();
      int right  = rect.right();

      // bypass fragment computation if there is no clipping
      if (bitmap.isScreenReset())
      {
         output.box(left, top, right, bottom, color);
         return this;
      }

      boolean corners = output.hasBoxBorderCorners();

      int width = right - left + 1;
      int height = bottom - top + 1;
      int corner = 0;

      if (corners)
      {
         width = width - 2;
         height = height - 2;
         corner = 1;
      }

      // upper-left corner
      if (corners && bitmap.canDraw(top, left))
      {
         output.drawULCorner(left, top, color);
      }

      // upper border
      if (width > 0)
      {
         drawHBorder(left + corner, top, width, color);
      }

      // upper-right corner
      if (corners && bitmap.canDraw(top, right))
      {
         output.drawURCorner(right, top, color);
      }

      // lower-left corner
      if (corners && bitmap.canDraw(bottom, left))
      {
         output.drawLLCorner(left, bottom, color);
      }

      // lower border
      if (width > 0)
      {
         drawHBorder(left + corner, bottom, width, color);
      }

      // lower-right corner
      if (corners && bitmap.canDraw(bottom, right))
      {
         output.drawLRCorner(right, bottom, color);
      }

      // vertical borders
      if (height > 0)
      {
         drawVBorder(left, top + corner, height, color);
         drawVBorder(right, top + corner, height, color);
      }
      
      return this;
   }

   /**
    * Draw box at interactive terminal.
    * 
    * @param   origin
    *          Upper left corner.
    * @param   size
    *          Box dimensions.
    * @param   color
    *          Box color.
    */
   public void boxToScreen(Point origin, Dimension size, Color color)
   {
      ScreenBitmap old = bitmap;
      
      try
      {
         // set output to screen
         bitmap = defaultOutput.getScreenBitmap();
      
         box(origin, size, color);
      }
      finally
      {
         // restore the bitmap
         bitmap = old;
      }
   }
   
   /**
    * Clear area occupied by specified widget.
    * 
    * @param   widget
    *          Reference widget. 
    */
   public void clear(Widget widget)
   {
      if (widget == null || isRedirected())
         return;

      clear(widget.screenLocation(),
            widget.dimension(),
            output.widgetColor(widget));
   }

   /**
    * Draw the upper or lower border.
    * 
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   width
    *          The border width.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   private void drawHBorder(int col, int row, int width, Color color)
   {
      // split the line into drawable fragments and draw them
      List<Point> frags = bitmap.getHorizontalFragments(row, col, width);
      for (Point frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         output.drawHBorder(col + start, row, length, color);
      }
   }

   /**
    * Draw the left or right border.
    * 
    * @param   col
    *          The column number where to draw.
    * @param   row
    *          The row number where to draw.
    * @param   height
    *          The border height.
    * @param   color
    *          The number of the color-pair (foreground+background)
    */
   private void drawVBorder(int col, int row, int height, Color color)
   {
      // split the line into drawable fragments and draw them
      List<Point> frags = null;
      if (isRedirected() == false)
      {
         frags = bitmap.getVerticalFragments(row, col, height);
      }
      else
      {
         frags = bitmap.getVerticalFragmentsAnyColumn(row, height);
      }

      for (Point frag : frags)
      {
         int start = frag.x;
         int end = frag.y;
         int length = end - start + 1;

         output.drawVBorder(col, row + start, length, color);
      }
   }

   /**
    * Get current cursor position.
    * 
    * @return  current cursor position.
    */
   public Point cursor()
   {
      return new Point(output.getX(), output.getY());
   }

   /**
    * Gets the number of columns on the screen.
    * 
    * @return  The number of columns (in characters).
    */
   public int screenWidth()
   {
      return output.screenWidth();
   }

   /**
    * Gets the number of rows on the screen.
    * 
    * @return  The number of rows (lines of characters).
    */
   public int screenHeight()
   {
      return output.screenHeight();
   }

   /**
    * Get screen dimensions.
    * 
    * @return screen dimension.
    */
   public Dimension screenDimension()
   {
      return new Dimension(screenWidth(), screenHeight());
   }

   /**
    * Reset the screen bitmap.
    * 
    * @return  <code>this</code> to enable chaining. 
    */
   public OutputManager resetBitmap()
   {
      bitmap.resetScreen();
      return this;
   }

   /**
    * Get a copy of the current screen bitmap and reset clipping entirely. All
    * screen positions become available for output. The previous state can be
    * restored using setBitmap() method.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getAndResetBitmap()
   {
      ScreenBitmap copy = bitmap.getCopy();
      bitmap.addRectangle(null);

      return copy;
   }

   /**
    * Get a copy of the current screen bitmap and clip output further so it is
    * guaranteed not to go outside of the provided rectangle.
    * 
    * @param   rect
    *          The extra clipping rectangle. <code>null</code> has no effect
    *          on clipping.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getAndSetBitmap(Rectangle rect)
   {
      ScreenBitmap copy = bitmap.getCopy();

      if (rect != null)
      {
         bitmap.clipRectangle(rect);
      }

      return copy;
   }

   /**
    * Get a copy of the current screen bitmap.
    * 
    * @return  a copy of the current screen bitmap.
    */
   public ScreenBitmap getBitmapCopy()
   {
      return bitmap.getCopy();
   }

   /**
    * Restore the bitmap to a previous state.
    * 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager setBitmap(ScreenBitmap bitmap)
   {
      this.bitmap = bitmap;
      output.setScreenBitmap(bitmap);
      return this;
   }

   /**
    * Mark a rectangle as needing repainting. Makes sense with the double
    * buffered output.
    * 
    * @param   rect
    *          rectangle to mark for repainting
    */
   public void refresh(Rectangle rect)
   {
      output.refresh(rect);
   }

   /**
    * Set cursor position.
    * 
    * @param   x
    *          Cursor X.
    * @param   y
    *          Cursor Y.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager at(int x, int y)
   {
      cursorInvalid = !output.cursorAt(x, y);
      return this;
   }

   /**
    * Set cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager at(Point p)
   {
      return at(p.x, p.y);
   }

   /**
    * Set real cursor position.
    * 
    * @param   p
    *          New cursor position.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager stayAt(Point p)
   {
      output.cursorStay(p.x, p.y);
      
      return this;
   }
   
   /**
    * Synchronize output with physical terminal.
    * 
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager sync()
   {
      output.sync();
      return this;
   }

   /**
    * Add a new clipping rectangle to the bitmap. If the rectangle is null,
    * fill the entire bitmap with <code>true</code> values. Else, if the
    * bitmap was just reset, fill it with <code>false</code> values and set
    * the clipping region to true.
    * 
    * @param   rect
    *          The clipping rectangle to be added.
    *          
    * @return  <code>this</code> to enable chaining.
    */
   public OutputManager clipWith(Rectangle rect)
   {
      if (rect != null && bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return this;
         }

         if (debug)
         {
            System.err.println("clipWith: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(rect);

         if (rect == null)
         {
            return this;
         }
      }

      bitmap.addRectangle(rect);
      return this;
   }

   /**
    * Fix the coordinates for the given rectangle. If any of the coordinates
    * is outside the screen, fix it. If the top/bottom or left/right are both
    * outside the screen, return <code>null</code>; this means the rectangle
    * should not be fixed, as it refers to an area outside the screen.
    * 
    * @param   rect
    *          The rectangle which needs its coordinates fixed.
    * 
    * @return  a rectangle instance with its coordinates fixed or
    *          <code>null</code> if the rectangle is referring to an area
    *          outside the screen.
    */
   private Rectangle fixCoordinates(Rectangle rect)
   {
      if (rect.left() < 0)
      {
         if (rect.right() >= 0)
         {
            rect = new Rectangle(rect.top(), 0, rect.bottom(), rect.right());
         }
         else
         {
            return null;
         }
      }

      if (rect.top() < 0)
      {
         if (rect.bottom() >= 0)
         {
            rect = new Rectangle(0, rect.left(), rect.bottom(), rect.right());
         }
         else
         {
            return null;
         }
      }

      // the bitmap.rows value is EXCLUSIVE (doesn't include the last row);
      // OTOH, the Rectangle._bottom/_top are INCLUSIVE.
      if (rect.bottom() >= bitmap.rows)
      {
         if (rect.top() < bitmap.rows())
         {
            rect = new Rectangle(rect.top(), rect.left(), bitmap.rows() - 1,
                                 rect.right());
         }
         else
         {
            return null;
         }
      }

      // the bitmap.cols value is EXCLUSIVE (doesn't include the last column);
      // OTOH, the Rectangle._left/_right are INCLUSIVE.
      if (rect.right() >= bitmap.cols())
      {
         if (rect.left() < bitmap.cols())
         {
            rect = new Rectangle(rect.top(), rect.left(), rect.bottom(),
                                 bitmap.cols - 1);
         }
         else
         {
            return null;
         }
      }

      return rect;
   }

   /**
    * Create a copy of this screen bitmap and count the invalid chars. This
    * initialization is needed to start searching for the frame from where to
    * start drawing.
    */
   public void startScreenSearch()
   {
      bitmap.startScreenSearch();
   }

   /**
    * Get the total number of available clipped characters, from the search
    * bitmap screen.
    * 
    * @return  The total number of invalid characters.
    */
   public int getInvalidChars()
   {
      return bitmap.getInvalidChars();
   }

   /**
    * Mark a certain area as searched and decrement the total number of
    * clipped characters.
    * 
    * @param   rect
    *          Area to mark as searched.
    * 
    * @return  <code>true</code> if this area contains a clipped character,
    *          which was not part of an upper widget.
    */
   public boolean markArea(Rectangle rect)
   {
      if (rect != null && bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return false;
         }

         if (debug)
         {
            System.err.println("markArea: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(rect);

         if (rect == null)
         {
            return false;
         }
      }

      return bitmap.markArea(rect);
   }

   /**
    * Check if the given location can draw data.
    * 
    * @param   rect
    *          The area of the widget which is checked for drawing.
    * 
    * @return  <code>true</code> if screen is reset or any of the bitmap
    *          locations within the given rectangle are <code>true</code>
    *          <code>false</code> if nothing can be drawn.
    */
   public boolean drawNeeded(Rectangle rect)
   {
      if (rect != null && bitmap.hasCoordinatesOutsideScreen(rect))
      {
         // ignore invalid requests
         if (bitmap.hasInvalidCoordinates(rect))
         {
            return false;
         }

         if (debug)
         {
            System.err.println("drawNeeded: received rectangle with incorrect" +
                               " coordinates: " + rect.toString());
         }

         rect = fixCoordinates(rect);

         if (rect == null)
         {
            return false;
         }
      }

      // bypass if there is no clipping
      if (bitmap.isScreenReset())
      {
         return true;
      }

      return bitmap.drawNeeded(rect);
   }

   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    */
   public void setInvalidate(boolean mode)
   {
      setInvalidate(mode, false);
   }

   /**
    * Sets the invalidation mode.
    * <p>
    * When invalidation mode is set to <code>true</code>, all
    * <code>PaintEvent</code>s are processed differently. Instead of causing
    * immediate drawing, they contribute their respective rectangles to the
    * invalidation rectangle. The moment the invalidation mode is being set to
    * <code>false</code>, the cumulative invalidation rectangle triggers
    * selective redrawing of the affected area.
    * 
    * @param   mode
    *          <code>true</code> is an opening bracket for the invalidation
    *          mode <code>false</code> is a closing bracket which
    *          conditionally flushes the accumulated paint requests
    * @param   flush
    *          only meaningful for mode <code>false</code>. requests
    *          unconditional flushing of all accumulated paint requests
    * 
    */
   public void setInvalidate(boolean mode, boolean flush)
   {
      if (mode)
      {
         invalidate++;
         return;
      }

      invalidate--;

      if (drawing)
         return;

      if ((invalidate == 0 || flush) && bitmap.hasRectangles())
      {
         drawing = true;

         Widget win = WindowManager.topWindow();

         win.draw();
         sync();

         bitmap.resetScreen();

         drawing = false;
      }
   }

   /**
    * Resets the invalidation mode. This method should be used exclusively as
    * a way of recovering after exceptions in the client before the control is
    * returned to the server. Such a recovery is required to keep the client
    * side drawing balanced.
    */
   public void resetInvalidate()
   {
      // setting the counter to 1 makes the call that follows reset it to 0
      // and draw the whole invalid rectangle
      invalidate = 1;
      setInvalidate(false, true);
   }

   /**
    * Returns and resets the invalidation mode. This method should be used for
    * activating triggers which, by definition, is the source of recursion.
    * 
    * @return  current invalidation level
    */
   public int saveInvalidate()
   {
      int save = invalidate;

      resetInvalidate();

      return save;
   }

   /**
    * Restores the invalidation mode. This method should be used for
    * activating triggers which, by definition, is the source of recursion.
    * 
    * @param   savedLevel
    *          saved current invalidation level
    */
   public void restoreInvalidate(int savedLevel)
   {
      invalidate = savedLevel;
   }

   /**
    * Queries the invalidation mode.
    * 
    * @return  <code>true</code> if invalidation is on
    */
   public boolean isInvalidate()
   {
      return invalidate > 0;
   }

   /**
    * @param colors
    */
   public void addColors(List<Color> colors)
   {
      driver.addColors(colors);
   }

   /**
    * Access driver-specific widget factory.
    * 
    * @return  reference to widget factory.
    */
   public WidgetFactory getFactory()
   {
      return driver.getFactory();
   }
}