/*
** Module   : BatchPrimitives.java
** Abstract : Output primitives for batch mode driver.
**
** Copyright (c) 2010-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131104 Created initial version.
*/

package com.goldencode.p2j.ui.client.chui.driver.batch;

import java.util.List;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

import com.goldencode.p2j.ui.client.chui.driver.console.*;

/**
 * Implementation of direct output primitives.
 * <p>
 * This implementation uses text console-based native driver for terminal access.
 * The specific of this implementation is that native API in many cases exposes
 * or is affected by NCURSES-specific details, such as key processing and 
 * special handling of pseudo-graphics characters. 
 */
public class BatchPrimitives
extends ConsolePrimitives
{
   /** Reference to console low-level driver. */
   private final BatchHelper bh;
   
   private boolean isActive = false;

   /**
    * Constructor.
    *
    * @param   helper
    *          ConsoleHelper instance used for direct drawing.
    */
   public BatchPrimitives(ConsoleHelper helper, BatchHelper batchHelper)
   {
//      super(helper);
      this.bh = batchHelper;
   }

   public void activate()
   {
      isActive = true;
   }
   
   public void deActivate()
   {
      isActive = false;
   }
   
   /**
    * Position the virtual cursor during the drawing.
    *
    * @param   x
    *          New cursor X.
    * @param   y
    *          New cursor Y.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      if (isActive)
      {
         isActive = false;
      }
//      else
//      {
//         isActive = true;
//      }

      return true;
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return  The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return 24;
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return  The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return 80;
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()
   {
      if (isActive)
      {
         isActive = false;
         bh.outputOperation();
      }
//      else
//      {
//         isActive = true;
//      }
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The screen container is safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      if (isActive)
      {
         isActive = false;
         bh.outputOperation();
      }
//      else
//      {
//         isActive = true;
//      }
   }
}
