/*
** Module   : ConsolePrimitives.java
** Abstract : Output primitives for text console-based driver.
**
** Copyright (c) 2010-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20130723 Created initial version
** 002 EVL 20130822 Adding new 6 characters into consideration to handle comb-box and
**                  selection-list widgets drawing.
** 003 EVL 20131104 Batch mode support.
*/

package com.goldencode.p2j.ui.client.chui.driver.console;

import java.util.List;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of direct output primitives.
 * <p>
 * This implementation uses text console-based native driver for terminal access.
 * The specific of this implementation is that native API in many cases exposes
 * or is affected by NCURSES-specific details, such as key processing and 
 * special handling of pseudo-graphics characters. 
 */
// EVL***class ConsolePrimitives
public class ConsolePrimitives // EVL***
extends DriverPrimitives
{
   /** Horizontal line drawing char. */
   private static int ACS_HLINE = 0;

   /** Vertical line drawing char. */
   private static int ACS_VLINE = 1;

   /** Upper left corner drawing char. */
   private static int ACS_ULCORNER = 2;

   /** Upper right corner drawing char. */
   private static int ACS_URCORNER = 3;

   /** Lower right corner drawing char. */
   private static int ACS_LRCORNER = 4;

   /** Lower left corner drawing char. */
   private static int ACS_LLCORNER = 5;

   /** Up arrow char for vertical scrollbars. */
   private static int ACS_UARROW = 6;

   /** Down arrow char for vertical scrollbars. */
   private static int ACS_DARROW = 7;

   /** Left arrow char for horizontal scrollbars. */
   private static int ACS_LARROW = 8;

   /** Right arrow char for horizontal scrollbars. */
   private static int ACS_RARROW = 9;

   /** Checker board char for scrollbars body. */
   private static int ACS_CKBOARD = 10;

   /** Current selection char for selection lists. */
   private static int ACS_SELECT = 11;

   /** Reference to console low-level driver. */
   private final ConsoleHelper helper;

   /** Normal attribute. */
   private int A_NORMAL;

   /** Underline attribute */
   private int A_UNDERLINE;

   /** Reverse attribute */
   private int A_REVERSE;

   /** Blink attribute */
   private int A_BLINK;

   /** Bold attribute */
   private int A_BOLD;

   /** Flag which controls use of colors. */
   private boolean useColor;

   // EVL***<<
   public ConsolePrimitives()
   {
      helper = null;
   }
   // EVL***>>
   /**
    * Constructor.
    *
    * @param   helper
    *          ConsoleHelper instance used for direct drawing.
    */
   public ConsolePrimitives(ConsoleHelper helper) // EVL***
// EVL***   ConsolePrimitives(ConsoleHelper helper)
   {
      this.helper = helper;

      A_NORMAL = ConsoleHelper.getAttribute(0);
      A_UNDERLINE = ConsoleHelper.getAttribute(2);
      A_REVERSE = ConsoleHelper.getAttribute(3);
      A_BLINK = ConsoleHelper.getAttribute(4);
      A_BOLD = ConsoleHelper.getAttribute(6);

      ACS_ULCORNER = ConsoleHelper.getACSchar(0);
      ACS_LLCORNER = ConsoleHelper.getACSchar(1);
      ACS_URCORNER = ConsoleHelper.getACSchar(2);
      ACS_LRCORNER = ConsoleHelper.getACSchar(3);
      ACS_HLINE = ConsoleHelper.getACSchar(8);
      ACS_VLINE = ConsoleHelper.getACSchar(9);
      ACS_CKBOARD = ConsoleHelper.getACSchar(14);
      ACS_UARROW = ConsoleHelper.getACSchar(18);
      ACS_DARROW = ConsoleHelper.getACSchar(19);
      ACS_LARROW = ConsoleHelper.getACSchar(20);
      ACS_RARROW = ConsoleHelper.getACSchar(21);
      ACS_SELECT = ConsoleHelper.getACSchar(22);

      useColor = helper.hasColors() && ConsoleHelper.isColorEnabled;
   }

   /**
    * Add a string to the virtual terminal at the current cursor position.
    *
    * @param   str
    *          The text to output.
    * @param   color
    *          The Color instance used to draw characters. Mapping into native
    *          colors is performed internally.
    */
   @Override
   public void append(String str, Color color)
   {
      // the input string is encoded as unicode which has 2 bytes for each
      // character, when anything other than 7-bit ASCII chars are used
      // the result is corrupted unless we convert to the native platform
      // charset (via getBytes()) --> as long as the native platform
      // character set is a single byte charset, then this result will
      // preserve the data properly all the way through to the terminal
      //TODO: use encoding
      byte[] bytes = str.getBytes();
      helper.addStringNative(bytes, bytes.length, attribute(color), color(color));
   }

   /**
    * Position the virtual cursor during the drawing.
    *
    * @param   x
    *          New cursor X.
    * @param   y
    *          New cursor Y.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorAt(int x, int y)
   {
      return helper.setCursorNative(x, y);
   }

   /**
    * Position the real cursor.
    *
    * @param   x
    *          New cursor X.
    * @param   y
    *          New cursor Y.
    *
    * @return  <code>true</code> if cursor positioning succeeded.
    */
   @Override
   public boolean cursorStay(int x, int y)
   {
      return helper.setCursorNative(x, y);
   }

   /**
    * Get absolute X position of the cursor.
    *
    * @return  The X position of the cursor.
    */
   @Override
   public int getX()
   {
      return helper.getx();
   }

   /**
    * Get absolute Y position of the cursor.
    *
    * @return  The Y position of the cursor.
    */
   @Override
   public int getY()
   {
      return helper.gety();
   }

   /**
    * Gets the number of rows on the screen.
    *
    * @return  The number of rows (lines of characters).
    */
   @Override
   public int screenHeight()
   {
      return helper.getScreenRowsNative();
   }

   /**
    * Gets the number of columns on the screen.
    *
    * @return  The number of columns (in characters).
    */
   @Override
   public int screenWidth()
   {
      return helper.getScreenColumnsNative();
   }

   /**
    * Synchronize the state of the physical terminal with the state of
    * the virtual terminal.
    */
   @Override
   public void sync()
   {
      helper.syncNative();
   }

   /**
    * Perform a bulk output operation for the whole screen at once with
    * output masking when desired. Set the cursor afterwards.
    * <p>
    * The screen container is safe for use with characters that are UNICODE
    * encoded which are also outside of the 7-bit ASCII range. Such
    * characters are passed through to the native layer directly and are
    * not converted to a natively encoded form.  This could be provided
    * but since one of the primary uses of this method is for drawing
    * characters (e.g. <code>ACS_HLINE</code>) which are encoded using the
    * full 32-bits, any attempt to convert these into a single byte native
    * encoding would lose data.
    * <p>
    * Because of the above limitation, DO NOT USE THIS FOR OUTPUT OF USER
    * DATA THAT COULD CONTAIN DATA OUTSIDE OF THE 7-bit ASCII RANGE.
    *
    * @param   screen
    *          Container for whole screen content and cursor position.
    */
   @Override
   public void updateScreen(ScreenData screen)
   {
      helper.addArrayNative(convertToBuffer(screen), screen.getCursorX(), screen.getCursorY());
   }

   /**
    * Extract attributes from given {@link Color} instance.
    *
    * @param   color
    *          Color.
    *
    * @return  Attributes recognized by native layer.
    */
   private int attribute(Color color)
   {
      int attr = A_NORMAL;

      if (color != null)
      {
         attr |= color.hasBlink() ? A_BLINK : 0;
         attr |= color.hasUnderline() ? A_UNDERLINE : 0;
         attr |= color.hasReverse() ? A_REVERSE : 0;
         attr |= color.hasBold() ? A_BOLD : 0;
      }

      return attr;
   }

   /**
    * Convert source character into form recognized by native driver.
    *
    * @param   character
    *          Input character.
    *
    * @return  encoded character.
    */
   private int character(char character)
   {
      // encode the change into the low level output matrix
      switch (character)
      {
         case OutputManager.UACS_HLINE:
            return ConsolePrimitives.ACS_HLINE;

         case OutputManager.UACS_VLINE:
            return ConsolePrimitives.ACS_VLINE;

         case OutputManager.UACS_ULCORNER:
            return ConsolePrimitives.ACS_ULCORNER;

         case OutputManager.UACS_URCORNER:
            return ConsolePrimitives.ACS_URCORNER;

         case OutputManager.UACS_LRCORNER:
            return ConsolePrimitives.ACS_LRCORNER;

         case OutputManager.UACS_LLCORNER:
            return ConsolePrimitives.ACS_LLCORNER;

         case 0x07F:
            return '^';

         case OutputManager.UACS_CKBOARD:
            return ConsolePrimitives.ACS_CKBOARD;
            
         case OutputManager.UACS_UARROW:
            return ConsolePrimitives.ACS_UARROW;
            
         case OutputManager.UACS_DARROW:
            return ConsolePrimitives.ACS_DARROW;
            
         case OutputManager.UACS_LARROW:
            return ConsolePrimitives.ACS_LARROW;
            
         case OutputManager.UACS_RARROW:
            return ConsolePrimitives.ACS_RARROW;
            
         case OutputManager.UACS_SELECT:
            return ConsolePrimitives.ACS_SELECT;
      }
      // convert UNICODE char to the output charset
      return Utils.toChar(character);
   }

   /**
    * Extract color from given {@link Color} instance.
    *
    * @param   color
    *          Color.
    *
    * @return  Color recognized by native layer.
    */
   private int color(Color color)
   {
      return (useColor && color != null) ? color.color() : 0;
   }

   /**
    * Convert screen data into array of integers recognized by native driver.
    *
    * @param   screen
    *          Screen data.
    *
    * @return  Array of encoded characters with colors and attributes.
    */
   private int[] convertToBuffer(ScreenData screen)
   {
      List<Cell> content = screen.list();
      int[] buffer = new int[content.size()];
      int i = 0;
      for (Cell cell : content)
      {
         buffer[i++] = encode(cell);
      }

      return buffer;
   }

   /**
    * Convert given cell into integer code recognized by native layer as
    * character, color and attributes.
    *
    * @param   cell
    *          Input cell.
    *
    * @return  Encoded character, color and attribute.
    */
   private int encode(Cell cell)
   {
      if (cell == null) {
         return -1;
      }

      return character(cell.character()) |
             attribute(cell.color()) |
             color(cell.color());
   }
}