/*
** Module   : UnstoppableExitException.java
** Abstract : represents a Progress condition (ERROR, END, STOP or QUIT)
**
** Copyright (c) 2005-2006, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20131115 Created initial version which represents a Progress uncatched condition.
*/

package com.goldencode.p2j.util;

/**
 * Represents a Progress condition (ERROR, END, STOP or QUIT).
 * <p>
 * This class is an unchecked exception because the generated code always
 * emits catch blocks regardless of whether the body of the <code>try</code>
 * block can generate these exceptions or not.  The Java compiler fails in 
 * this condition (we have too much exception processing). The other solution
 * to this is to exhaustively analyze all <code>try</code> bodies to
 * determine if any <code>catch</code> blocks must be removed.  This has no
 * runtime benefit and is such a huge/complex effort that it is not
 * warrented.  Instead we use unchecked exceptions.
 */
public class UnstoppableExitException
extends RuntimeException
{
   /**
    * Constructor which accepts an error message.
    *
    * @param   message
    *          Error message.
    */
   public UnstoppableExitException(String message)
   {
      super(message);
   }
    
   /**
    * Constructor which accepts an error message and a root cause. This
    * variant is used when the error condition is detected as the result of
    * catching another exception.
    *
    * @param   message
    *          Error message.
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public UnstoppableExitException(String message, Throwable rootCause)
   {
      super(message, rootCause);
   }
    
   /**
    * Constructor which accepts and wrappers a root cause. This variant is
    * used when the error condition is detected as the result of catching
    * another exception, and the message or context of the root cause is
    * sufficient to describe the problem.
    *
    * @param   rootCause
    *          Exception which has been caught in the course of processing.
    */
   public UnstoppableExitException(Throwable rootCause)
   {
      super(rootCause);
   }
}
